import numpy
from primordial.equations.cosmology import Equations as _Equations

class Equations(_Equations):
    """ Cosmology equations in time 

    Solves background variables in cosmic time for curved and flat universes
    using the Friedmann equation.

    Independent variable:
        N: efolds
    
    Variables:
        t: cosmic time
    
    """
    def __init__(self, H0, Omega_r, Omega_m, Omega_k, Omega_l):
        super(Equations, self).__init__(H0, Omega_r, Omega_m, Omega_k, Omega_l)

        self.add_independent_variable('N')
        self.add_variable('t')

    def __call__(self, N, y):
        """ The derivative function for underlying variables,
            computed using the Klein-Gordon equation """
        dy = numpy.zeros_like(y)
        dy[self['t']] = 1/self.H(N, y)
        return dy


class initial_conditions(object):
    def __init__(self, Ni):
        self.t0 = Ni

    def __call__(self, equations, y0):
        y0[equations['t']] = 0



from numpy.testing import assert_allclose, assert_almost_equal
from primordial.equations.t.inflation import Equations as Eqs_t, Inflation_start_initial_conditions as IC_t
from primordial.equations.N.inflation import Equations as Eqs_N, Inflation_start_initial_conditions as IC_N

from primordial.equations.events import Inflation
from primordial.equations.inflation_potentials import ChaoticPotential
from primordial.solver import solve

def test_inflation():
    atol, rtol = 1e-8, 1e-8

    V = ChaoticPotential(m=1)
    N_0, phi_0 = 0, 17

    for K in [-1, 0, 1]:
        # Solve in N
        equations = Eqs_N(K, V)
        ic = IC_N(N_0, phi_0)
        events = Inflation(equations, direction=-1, terminal=True)
        sol_N = solve(equations, ic, events=events, atol=atol*1e-1, rtol=rtol*1e-1)

        N = sol_N.N[1:-1]
        t = sol_N.t(N)

        # Solve in t
        equations = Eqs_t(K, V)
        ic = IC_t(N_0, phi_0)
        events = Inflation(equations, direction=-1, terminal=True)
        sol_t = solve(equations, ic, events=events, t_eval=t, atol=atol*1e-1, rtol=rtol*1e-1)

        # Check t-N consistency
        assert_allclose(sol_t.N(t), N, atol, rtol)
        assert_allclose(t, sol_N.t(N), atol, rtol)

        # Check phi
        assert_allclose(sol_t.phi(t), sol_N.phi(N), atol, rtol)

        # Check H
        assert_allclose(sol_N.H(N), sol_t.H(t), atol, rtol)



import numpy
from primordial.equations.inflation import Equations as _Equations

class Equations(_Equations):
    """ Background equations in time 

    Solves bacgkround variables in cosmic time for curved and flat universes
    using the Klein-Gordon and Friedmann equations.

    Independent variable:
        N: e-folds (log a)
    
    Variables:
        phi: inflaton field
        dphi: d/dN (phi) 
        t: cosmic time
    
    """
    def __init__(self, K, potential):
        super(Equations, self).__init__(K, potential)
        self.add_independent_variable('N')
        self.add_variables(['phi', 'dphi', 't'])

    def __call__(self, N, y):
        """ The derivative function for underlying variables,
            computed using the Klein-Gordon equation """
        H2 = self.H2(N, y)
        dphi = self.dphi(N, y)
        dVdphi = self.dVdphi(N, y)
        dlogH = self.dlogH(N, y)
        ddphi = -(dlogH + 3)* dphi - dVdphi/H2

        dy = numpy.zeros_like(y)
        dy[self['phi']] = dphi
        dy[self['dphi']] = ddphi
        dy[self['t']] = 1./self.H(N, y)

        return dy

    def H2(self, N, y):
        """ The square of the Hubble parameter,
            computed using the Friedmann equation """
        V = self.V(N, y)
        dphi = self.dphi(N, y) 

        return (2*V - 6*self.K*numpy.exp(-2*N)) / (6. - dphi**2)

    def dlogH(self, N, y):
        """ d/dN log H """
        dphi = self.dphi(N, y) 
        H2 = self.H2(N, y) 
        return - dphi**2/2. + self.K * numpy.exp(-2*N)/H2

    def inflating(self, N, y):
        """ Inflation diagnostic """
        return self.V(N, y) / (self.V(N, y)/2. - self.K*numpy.exp(-2*N)) - self.dphi(N, y)**2 


class Inflation_start_initial_conditions(object):
    def __init__(self, N_e, phi_e):
        self.t0 = N_e
        self.phi_e = phi_e

    def __call__(self, equations, y0):
        V = equations.potential(self.phi_e)
        N = self.t0
        y0[equations['phi']] = self.phi_e
        y0[equations['dphi']] = -numpy.sqrt(V / (V/2. - equations.K*numpy.exp(-2*N)))
        y0[equations['t']] = 0.



import numpy

class Event(object):
    """ Base class for events.

    Gives a more usable wrapper to 

    Derived classes must define:

    __call__(self, t, y)
        Scalar root function for determining event

    Attributes:
    ----------

    equations: Equations
        The equations for computing derived variables.

    direction: [-1, 0, +1]
        The direction of the root finding (if any)

    terminal: bool
        Whether to stop at this root
    """
    def __init__(self, equations, direction=0, terminal=False, value=0):
        self.equations = equations
        self.direction = direction
        self.terminal = terminal
        self.value = value


class Inflation(Event):
    """ Inflation entry/exit """
    def __call__(self, t, y):
        return self.equations.inflating(t, y) - self.value


class Collapse(Event):
    """ Tests if H^2 is positive """
    def __call__(self, t, y):
        return self.equations.H2(t, y) - self.value

class ModeExit(Event):
    """ When mode exits the horizon """
    def __call__(self, t, y):
        return numpy.log(numpy.abs(self.equations.H2(t, y)))/2.+self.equations.N(t,y) - numpy.log(self.value)

class UntilN(Event):
    """ Stop at N """
    def __call__(self, t, y):
        return self.equations.N(t, y) - self.value



from numpy.testing import assert_allclose, assert_almost_equal
from primordial.equations.t.mukhanov_sasaki import Equations as Eqs_t, Inflation_start_initial_conditions as IC_t
from primordial.equations.N.mukhanov_sasaki import Equations as Eqs_N, Inflation_start_initial_conditions as IC_N

from primordial.equations.events import Inflation
from primordial.equations.inflation_potentials import ChaoticPotential
from primordial.solver import solve
import numpy

def test_mukhanov_sasaki():
    atol, rtol = 1e-8, 1e-8

    V = ChaoticPotential(m=1)
    N_0, phi_0 = 0, 17

    for k in 2**numpy.arange(7):
        for K in [-1, 0, 1]:
            # Solve in N
            equations = Eqs_N(K, V, k)
            ic = IC_N(N_0, phi_0)
            events = Inflation(equations, direction=-1, terminal=True)
            sol_N = solve(equations, ic, events=events, atol=atol*1e-1, rtol=rtol*1e-1)

            N = sol_N.N[1:-1]
            t = sol_N.t(N)

            # Solve in t
            equations = Eqs_t(K, V, k)
            ic = IC_t(N_0, phi_0)
            events = Inflation(equations, direction=-1, terminal=True)
            sol_t = solve(equations, ic, events=events, t_eval=t, atol=atol*1e-1, rtol=rtol*1e-1)

            # Check R2
            assert_allclose(sol_N.R2(N), sol_t.R2(t), atol, rtol)
            assert_allclose(sol_N.dR2(N)*sol_N.H(N), sol_t.dR2(t), atol, rtol)




import numpy
from scipy import integrate

def solve(equations, ic, interp1d_kwargs={}, *args, **kwargs):
    y0 = numpy.zeros(len(equations))
    ic(equations, y0)
    sol = integrate.solve_ivp(equations, (ic.t0, 1e300), y0, *args, **kwargs)
    return equations.sol(sol, **interp1d_kwargs)




import numpy
from primordial.equations.inflation import Equations as _Equations

class Equations(_Equations):
    """ Background equations in time 

    Solves bacgkround variables in cosmic time for curved and flat universes
    using the Klein-Gordon and Friedmann equations.

    Independent variable:
        t: cosmic time
    
    Variables:
        N: efolds
        phi: inflaton field
        dphi: d (phi) / dt
    
    """
    def __init__(self, K, potential):
        super(Equations, self).__init__(K, potential)
        self.add_independent_variable('t')
        self.add_variables(['N', 'phi', 'dphi'])

    def __call__(self, t, y):
        """ The derivative function for underlying variables,
            computed using the Klein-Gordon equation """
        H = self.H(t, y)
        dphi = self.dphi(t, y)
        dVdphi = self.dVdphi(t, y)
        ddphi = -3*H*dphi - dVdphi

        dy = numpy.zeros_like(y)
        dy[self['N']] = H
        dy[self['phi']] = dphi
        dy[self['dphi']] = ddphi

        return dy

    def H2(self, t, y):
        """ The square of the Hubble parameter,
            computed using the Friedmann equation """
        N = self.N(t, y)
        V = self.V(t, y)
        dphi = self.dphi(t, y) 
        N = self.N(t, y) 
        return (dphi**2/2. + V)/3. - self.K*numpy.exp(-2*N)

    def inflating(self, t, y):
        """ Inflation diagnostic """
        return self.V(t, y) - self.dphi(t, y)**2 


class KD_initial_conditions(object):
    def __init__(self, t0, N_p, phi_p):
        self.t0 = t0
        self.N_p = N_p
        self.phi_p = phi_p

    def __call__(self, equations, y0):
        t0 = self.t0
        b = equations.K * numpy.exp(-2*self.N_p)
        y0[equations['N']] = self.N_p + numpy.log(t0)/3 - 9./14 * b * t0**(4./3)
        y0[equations['phi']] = self.phi_p - numpy.sqrt(2./3)*numpy.log(t0) - 27*numpy.sqrt(6)/56*b*t0**(4./3)
        y0[equations['dphi']] = -numpy.sqrt(2./3)/t0 - 9*numpy.sqrt(6)/14 * b * t0**(1./3)


class Inflation_start_initial_conditions(object):
    def __init__(self, N_e, phi_e):
        self.t0 = 0.
        self.N_e = N_e
        self.phi_e = phi_e

    def __call__(self, equations, y0):
        y0[equations['N']] = self.N_e
        y0[equations['phi']] = self.phi_e
        y0[equations['dphi']] = -numpy.sqrt(equations.potential(self.phi_e))



import numpy
from primordial.equations.t.inflation import Equations as BackgroundEquations, KD_initial_conditions as BackgroundKD_initial_conditions, Inflation_start_initial_conditions as BackgroundInflation_start_initial_conditions

class Equations(BackgroundEquations):
    def __init__(self, K, potential, k):
        super(Equations, self).__init__(K, potential)
        self.k = k
        self.add_variables(['R1', 'dR1', 'R2', 'dR2'])


    def __call__(self, t, y):
        """ The derivative function for underlying variables,
            computed using the Mukhanov-Sasaki equation equation """

        # Compute background variables
        dy = super(Equations, self).__call__(t, y)

        # Get useful variables
        N = self.N(t, y)
        H = self.H(t, y)
        H2 = self.H2(t, y)
        dphi = self.dphi(t, y)
        ddphi = dy[self['dphi']]
        R1 = self.R1(t, y)
        R2 = self.R2(t, y)
        dR1 = self.dR1(t, y)
        dR2 = self.dR2(t, y)

        k = self.k
        K = self.K
        if self.K > 0:
            k2 = k*(k+2)-3*K
        else:
            k2 = k**2-3*K

        E = dphi**2 / H2 / 2.
        alpha = E*K + k2
        gamma = (k2**2 - k2*K - E * k2*K - 2*k2*ddphi*K/dphi/H - E*K**2 + 2*k2*K**2/H2*numpy.exp(-2*N))*numpy.exp(-2*N)
        beta = (3*H*k2 + 2*E*H*k2 + 2*k2*ddphi/dphi + 3*E*H*K - 2*k2*K/H*numpy.exp(-2*N))

        ddR1 = -(beta*dR1 + gamma*R1)/alpha
        ddR2 = -(beta*dR2 + gamma*R2)/alpha

        dy[self['R1']] = dR1
        dy[self['dR1']] = ddR1
        dy[self['R2']] = dR2
        dy[self['dR2']] = ddR2

        return dy

class KD_initial_conditions(BackgroundKD_initial_conditions):
    def __call__(self, equations, y0):
        super(KD_initial_conditions, self).__call__(equations, y0)
        y0[equations['R1']] = 0
        y0[equations['dR1']] = equations.k
        y0[equations['R2']] = 1
        y0[equations['dR2']] = 0


class Inflation_start_initial_conditions(BackgroundInflation_start_initial_conditions):
    def __call__(self, equations, y0):
        super(Inflation_start_initial_conditions, self).__call__(equations, y0)
        y0[equations['R1']] = 0
        y0[equations['dR1']] = equations.k
        y0[equations['R2']] = 1
        y0[equations['dR2']] = 0



""" Base classes for inflationary solvers"""
import numpy
from primordial.equations.equations import Equations as _Equations

class Equations(_Equations):

    def __init__(self, K, potential):
        super(Equations, self).__init__()
        self.potential = potential
        self.K = K

    def H(self, t, y):
        """ Hubble parameter"""
        return numpy.sqrt(self.H2(t, y))

    def V(self, t, y):
        """ Potential """
        return self.potential(self.phi(t, y))

    def dVdphi(self, t, y):
        """ Potential derivative """
        return self.potential.d(self.phi(t, y))

    def sol(self, sol, **kwargs):
        """ Post-process solution of solve_ivp """
        sol.H = self.interp1d(sol.t, self.H(sol.t, sol.y), **kwargs)
        sol = super(Equations, self).sol(sol, **kwargs)
        return sol



from numpy.testing import assert_allclose, assert_almost_equal
from primordial.equations.t.cosmology import Equations as Eqs_t, initial_conditions as IC_t
from primordial.equations.N.cosmology import Equations as Eqs_N, initial_conditions as IC_N

from primordial.equations.events import UntilN
from primordial.solver import solve
from primordial.units import H0_h, tp

def test_cosmology():
    atol, rtol = 1e-8, 1e-8

    h = 0.6732
    Omega_b = 0.022383     / h**2
    Omega_m = 0.14314      / h**2
    Omega_r = 4.18343e-5   / h**2 # Assumes T = 2.7255, and includes neutrinos
    H0 = H0_h * h

    for Omega_k in [-0.1, 0, 0.1]:
        Omega_l = 1-Omega_m-Omega_r-Omega_k

        # Solve in N
        equations = Eqs_N(H0, Omega_r, Omega_m, Omega_k, Omega_l)

        Ni = 70 if Omega_k else -70
        ic = IC_N(Ni)

        Ne = equations.N0 if Omega_k else 0
        events = UntilN(equations, terminal=True, value=Ne)
        sol_N = solve(equations, ic, events=events, atol=atol*1e-1, rtol=rtol*1e-1)

        N = sol_N.N[1:-1]
        t = sol_N.t(N)

        # Solve in t
        equations = Eqs_t(H0, Omega_r, Omega_m, Omega_k, Omega_l)
        ic = IC_t(Ni)
        events = UntilN(equations, terminal=True, value=Ne)
        sol_t = solve(equations, ic, events=events, t_eval=t, atol=atol*1e-1, rtol=rtol*1e-1)

        ## Check t-N consistency
        assert_allclose(sol_t.N(t), N, atol, rtol)
        assert_allclose(t, sol_N.t(N), atol, rtol)

        # Check H
        assert_allclose(sol_N.H(N), sol_t.H(t), atol*10, rtol*10)



import numpy
from primordial.equations.equations import Equations as _Equations
from scipy.interpolate import interp1d

class Equations(_Equations):
    """ Cosmology equations

    Solves background variables in cosmic time for curved and flat universes
    using the Friedmann equation.

    Independent variable:
        N: efolds
    
    Variables:
        t: cosmic time
    
    """
    def __init__(self, H0, Omega_r, Omega_m, Omega_k, Omega_l):
        self.H0 = H0
        Omega = Omega_r + Omega_m + Omega_k + Omega_l
        self.Omega_m = Omega_m/Omega
        self.Omega_l = Omega_l/Omega
        self.Omega_k = Omega_k/Omega
        self.Omega_r = Omega_r/Omega
        if self.Omega_k==0:
            self.N0 = 0
        else:
            self.N0 = numpy.log(1./H0/numpy.sqrt(numpy.abs(self.Omega_k)))

    def H(self, t, y):
        """ Hubble parameter"""
        return numpy.sqrt(self.H2(t, y))

    def H2(self, t, y):
        """ The square of the Hubble parameter,
            computed using the Friedmann equation """
        N = self.N(t, y)
        return self.H0**2 * (
                  self.Omega_r * numpy.exp(4*(self.N0-N)) 
                + self.Omega_m * numpy.exp(3*(self.N0-N))
                + self.Omega_k * numpy.exp(2*(self.N0-N))
                + self.Omega_l
                )

    def sol(self, sol, **kwargs):
        """ Post-process solution of solve_ivp """
        sol.H = self.interp1d(sol.t, self.H(sol.t, sol.y), **kwargs)
        sol = super(Equations, self).sol(sol, **kwargs)
        return sol



import numpy
from scipy.constants import c, hbar, G

lp = numpy.sqrt(hbar*G/c**3)
mp = numpy.sqrt(hbar*c/G) 
tp = numpy.sqrt(hbar*G/c**5) 
Mpc = 3.086e+22
H0_h = 100 *1000 / Mpc * tp
DA = 13.91726e3
kstar = 0.05 * Mpc**-1 /( lp**2 / tp)



import numpy
import scipy.interpolate
from types import MethodType

class Equations(dict):
    """ Base class for equations.

    Allows one to compute derivatives and derived variables.
    Most of the other classes take 'equations' as an object. 

    Derived classes must define:

    __call__(self, t, y)
        The derivative function for underlying variables

    and add variable names using ``add_variable`` or ``add_variables``

    """

    def sol(self, sol, **kwargs):
        t, j = numpy.unique(sol.t,return_index=True)
        del sol.t
        for name, i in self.items():
            setattr(sol, name, self.interp1d(t, sol.y[i, j], **kwargs))
        tt = self.independent_variable
        setattr(sol, tt + '_events', sol.pop('t_events'))
        setattr(sol, tt, t)
        return sol

    def interp1d(self, x, y, **kwargs):
        kind = kwargs.pop('kind', 'cubic')
        bounds_error = kwargs.pop('bounds_error', False)
        return scipy.interpolate.interp1d(x, y, kind=kind, bounds_error=bounds_error, **kwargs)

    def add_independent_variable(self, name):
        """ Define the name of the independent variable """
        def method(self, t, y):
            return t
        setattr(self, name, MethodType(method, self))
        self.independent_variable = name

    def add_variable(self, name):
        """ Add variable to equations.
        
        * creates an index for the location of variable in y
        * creates a class method of the same name with signature
          name(self, t, y) that should be used to extract the variable value in
          an index-independent manner.
        """
        self[name] = len(self)
        def method(self, t, y):
            return numpy.array(y)[self[name],...]
        setattr(self, name, MethodType(method, self))

    def add_variables(self, names):
        """ Add multiple variables to the equations """
        for name in names:
            self.add_variable(name)

    @property
    def variables(self):
        return list(val for val in self.values())
